print("import: 'dulwich'")
import dulwich

print("import: 'dulwich.tests'")
import dulwich.tests

print("import: 'dulwich.tests.compat'")
import dulwich.tests.compat

print("import: 'dulwich.tests.utils'")
import dulwich.tests.utils

print("import: 'dulwich.archive'")
import dulwich.archive

print("import: 'dulwich.client'")
import dulwich.client

print("import: 'dulwich.config'")
import dulwich.config

print("import: 'dulwich.diff_tree'")
import dulwich.diff_tree

print("import: 'dulwich.errors'")
import dulwich.errors

print("import: 'dulwich.file'")
import dulwich.file

print("import: 'dulwich.hooks'")
import dulwich.hooks

print("import: 'dulwich.ignore'")
import dulwich.ignore

print("import: 'dulwich.index'")
import dulwich.index

print("import: 'dulwich.object_store'")
import dulwich.object_store

print("import: 'dulwich.objects'")
import dulwich.objects

print("import: 'dulwich.objectspec'")
import dulwich.objectspec

print("import: 'dulwich.pack'")
import dulwich.pack

print("import: 'dulwich.patch'")
import dulwich.patch

print("import: 'dulwich.protocol'")
import dulwich.protocol

print("import: 'dulwich.reflog'")
import dulwich.reflog

print("import: 'dulwich.refs'")
import dulwich.refs

print("import: 'dulwich.repo'")
import dulwich.repo

print("import: 'dulwich.server'")
import dulwich.server

print("import: 'dulwich.walk'")
import dulwich.walk

print("import: 'dulwich.web'")
import dulwich.web

